@@ -15,16 +15,17 @@
 #    under the License.
 
 import collections
-from concurrent import futures
 import re
-import threading
 import time
 
+import eventlet
+from eventlet import tpool
 from os_win import exceptions
 from os_win import utilsfactory
+from oslo_concurrency import lockutils
 from oslo_config import cfg
 from oslo_log import log as logging
-import six
+import threading
 
 from hyperv.common.i18n import _, _LE, _LW, _LI  # noqa
 from hyperv.neutron import _common_utils as c_util
@@ -37,6 +38,7 @@ CONF.import_group('NVGRE', 'hyperv.neutron.config')
 LOG = logging.getLogger(__name__)
 
 _port_synchronized = c_util.get_port_synchronized_decorator('n-hv-agent-')
+synchronized = lockutils.synchronized_with_prefix('n-hv-agent-')
 
 
 class HyperVNeutronAgentMixin(object):
@@ -70,6 +72,7 @@ networking-plugin-hyperv_agent.html
         self._port_metric_retries = {}
 
         self._nvgre_enabled = False
+        self._cache_lock = threading.Lock()
 
         conf = conf or {}
         agent_conf = conf.get('AGENT', {})
@@ -90,9 +93,9 @@ networking-plugin-hyperv_agent.html
         self.enable_security_groups = security_conf.get(
             'enable_security_group', False)
 
+        tpool.set_num_threads(self._worker_count)
+
         self._load_physical_network_mappings(self._phys_net_map)
-        self._init_nvgre()
-        self._workers = futures.ThreadPoolExecutor(self._worker_count)
 
     def _load_physical_network_mappings(self, phys_net_vswitch_mappings):
         self._physical_network_mappings = collections.OrderedDict()
@@ -106,7 +109,7 @@ networking-plugin-hyperv_agent.html
                 vswitch = parts[1].strip()
                 self._physical_network_mappings[pattern] = vswitch
 
-    def _init_nvgre(self):
+    def init_nvgre(self):
         # if NVGRE is enabled, self._nvgre_ops is required in order to properly
         # set the agent state (see get_agent_configrations method).
 
@@ -139,7 +142,7 @@ networking-plugin-hyperv_agent.html
         return phys_network_name
 
     def _get_network_vswitch_map_by_port_id(self, port_id):
-        for network_id, map in six.iteritems(self._network_vswitch_map):
+        for network_id, map in self._network_vswitch_map.items():
             if port_id in map['ports']:
                 return (network_id, map)
 
@@ -247,12 +250,7 @@ networking-plugin-hyperv_agent.html
         self._utils.connect_vnic_to_vswitch(map['vswitch_name'], port_id)
 
         if network_type == constants.TYPE_VLAN:
-            LOG.info(_LI('Binding VLAN ID %(segmentation_id)s '
-                         'to switch port %(port_id)s'),
-                     dict(segmentation_id=segmentation_id, port_id=port_id))
-            self._utils.set_vswitch_port_vlan_id(
-                segmentation_id,
-                port_id)
+            self._vlan_driver.bind_vlan_port(port_id, segmentation_id)
         elif network_type == constants.TYPE_NVGRE and self._nvgre_enabled:
             self._nvgre_ops.bind_nvgre_port(
                 segmentation_id, map['vswitch_name'], port_id)
@@ -335,12 +333,11 @@ networking-plugin-hyperv_agent.html
                                  device_details['physical_network'],
                                  device_details['segmentation_id'],
                                  device_details['admin_state_up'])
+            if CONF.AGENT.enable_qos_extension:
+                self._qos_ext.handle_port(self.context, device_details)
 
-            LOG.debug("Updating port %s status as UP.", port_id)
-            self.plugin_rpc.update_device_up(self.context,
-                                             device,
-                                             self.agent_id,
-                                             self._host)
+            LOG.debug("Updating cached port %s status as UP.", port_id)
+            self._update_port_status_cache(device, device_bound=True)
             LOG.info("Port %s processed.", port_id)
         except Exception:
             LOG.exception(_LE("Exception encountered while processing port "
@@ -369,31 +366,25 @@ networking-plugin-hyperv_agent.html
                              "%(device_details)s"),
                          {'device': device, 'device_details': device_details})
 
-                self._workers.submit(self._process_added_port, device_details)
+                eventlet.spawn_n(self._process_added_port, device_details)
 
             # remove the port from added ports set, so it doesn't get
             # reprocessed.
             self._added_ports.discard(device)
 
     def _treat_devices_removed(self):
-        for device in list(self._removed_ports):
-            LOG.info(_LI("Removing port %s"), device)
-            try:
-                self.plugin_rpc.update_device_down(self.context,
-                                                   device,
-                                                   self.agent_id,
-                                                   self._host)
-            except Exception as e:
-                LOG.debug("Removing port failed for device %(device)s: %(e)s",
-                          dict(device=device, e=e))
-                continue
+        for device in self._removed_ports.copy():
+            eventlet.spawn_n(self._process_removed_port, device)
 
-            self._port_unbound(device, vnic_deleted=True)
-            self.sec_groups_agent.remove_devices_filter([device])
+    def _process_removed_port(self, device):
+        self._update_port_status_cache(device, device_bound=False)
 
-            # if the port unbind was successful, remove the port from removed
-            # set, so it won't be reprocessed.
-            self._removed_ports.discard(device)
+        self._port_unbound(device, vnic_deleted=True)
+        self.sec_groups_agent.remove_devices_filter([device])
+
+        # if the port unbind was successful, remove the port from removed
+        # set, so it won't be reprocessed.
+        self._removed_ports.discard(device)
 
     def _process_added_port_event(self, port_name):
         LOG.info(_LI("Hyper-V VM vNIC added: %s"), port_name)
@@ -410,12 +401,44 @@ networking-plugin-hyperv_agent.html
 
         for event_type, callback in event_callback_pairs:
             listener = self._utils.get_vnic_event_listener(event_type)
-            thread = threading.Thread(target=listener, args=(callback,))
-            thread.start()
+            eventlet.spawn_n(listener, callback)
+
+    def _update_port_status_cache(self, device, device_bound=True):
+        with self._cache_lock:
+            if device_bound:
+                self._bound_ports.add(device)
+                self._unbound_ports.discard(device)
+            else:
+                self._bound_ports.discard(device)
+                self._unbound_ports.add(device)
+
+    @synchronized('n-plugin-notifier')
+    def _notify_plugin_on_port_updates(self):
+        if not (self._bound_ports or self._unbound_ports):
+            return
+
+        with self._cache_lock:
+            bound_ports = self._bound_ports.copy()
+            unbound_ports = self._unbound_ports.copy()
+
+        self.plugin_rpc.update_device_list(self.context,
+                                           list(bound_ports),
+                                           list(unbound_ports),
+                                           self.agent_id,
+                                           self._host)
+
+        with self._cache_lock:
+            self._bound_ports = self._bound_ports.difference(bound_ports)
+            self._unbound_ports = self._unbound_ports.difference(
+                unbound_ports)
 
     def daemon_loop(self):
+        # The following sets contain ports that are to be processed.
         self._added_ports = self._utils.get_vnic_ids()
         self._removed_ports = set()
+        # The following sets contain ports that have been processed.
+        self._bound_ports = set()
+        self._unbound_ports = set()
 
         self._create_event_listeners()
 
@@ -423,6 +446,8 @@ networking-plugin-hyperv_agent.html
             try:
                 start = time.time()
 
+                eventlet.spawn_n(self._notify_plugin_on_port_updates)
+
                 # notify plugin about port deltas
                 if self._added_ports:
                     LOG.debug("Agent loop has new devices!")
