@@ -18,12 +18,8 @@
 Unit tests for Windows Hyper-V virtual switch neutron driver
 """
 
-from concurrent import futures
-import time
-
 import mock
 from os_win import exceptions
-from os_win import utilsfactory
 from oslo_config import cfg
 
 from hyperv.neutron import constants
@@ -34,17 +30,15 @@ from hyperv.tests import base
 CONF = cfg.CONF
 
 
-class TestHyperVNeutronAgent(base.BaseTestCase):
+class TestHyperVNeutronAgent(base.HyperVBaseTestCase):
 
     _FAKE_PORT_ID = 'fake_port_id'
 
     def setUp(self):
         super(TestHyperVNeutronAgent, self).setUp()
-        utilsfactory_patcher = mock.patch.object(utilsfactory, '_get_class')
-        utilsfactory_patcher.start()
-        self.addCleanup(utilsfactory_patcher.stop)
 
         self.agent = hyperv_neutron_agent.HyperVNeutronAgentMixin()
+        self.agent._qos_ext = mock.MagicMock()
         self.agent.plugin_rpc = mock.Mock()
         self.agent._metricsutils = mock.MagicMock()
         self.agent._utils = mock.MagicMock()
@@ -56,7 +50,7 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
         self.agent.notifier = mock.Mock()
         self.agent._utils = mock.MagicMock()
         self.agent._nvgre_ops = mock.MagicMock()
-        self.agent._workers = mock.MagicMock()
+        self.agent._vlan_driver = mock.MagicMock()
 
     def test_load_physical_network_mappings(self):
         test_mappings = ['fakenetwork1:fake_vswitch',
@@ -120,7 +114,7 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
 
     @mock.patch.object(hyperv_neutron_agent.nvgre_ops, 'HyperVNvgreOps')
     def test_init_nvgre_disabled(self, mock_hyperv_nvgre_ops):
-        self.agent._init_nvgre()
+        self.agent.init_nvgre()
         self.assertFalse(mock_hyperv_nvgre_ops.called)
         self.assertFalse(self.agent._nvgre_enabled)
 
@@ -128,14 +122,14 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
     def test_init_nvgre_no_tunnel_ip(self, mock_hyperv_nvgre_ops):
         self.config(enable_support=True, group='NVGRE')
         self.assertRaises(exception.NetworkingHyperVException,
-                          self.agent._init_nvgre)
+                          self.agent.init_nvgre)
 
     @mock.patch.object(hyperv_neutron_agent.nvgre_ops, 'HyperVNvgreOps')
     def test_init_nvgre_enabled(self, mock_hyperv_nvgre_ops):
         self.config(enable_support=True, group='NVGRE')
         self.config(provider_tunnel_ip=mock.sentinel.tunneling_ip,
                     group='NVGRE')
-        self.agent._init_nvgre()
+        self.agent.init_nvgre()
         mock_hyperv_nvgre_ops.assert_called_once_with(
             list(self.agent._physical_network_mappings.values()))
 
@@ -341,9 +335,7 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
 
     @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
                        '_provision_network')
-    def test_port_bound_nvgre(self, mock_provision_network):
-        self.agent._nvgre_enabled = True
-        network_type = constants.TYPE_NVGRE
+    def _check_port_bound_net_type(self, mock_provision_network, network_type):
         net_uuid = 'my-net-uuid'
         fake_map = {'vswitch_name': mock.sentinel.vswitch_name,
                     'ports': []}
@@ -363,6 +355,17 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
             mock.sentinel.physical_network, mock.sentinel.segmentation_id)
         self.agent._utils.connect_vnic_to_vswitch.assert_called_once_with(
             mock.sentinel.vswitch_name, mock.sentinel.port_id)
+
+    def test_port_bound_vlan(self):
+        self._check_port_bound_net_type(network_type=constants.TYPE_VLAN)
+
+        self.agent._vlan_driver.bind_vlan_port.assert_called_once_with(
+            mock.sentinel.port_id, mock.sentinel.segmentation_id)
+
+    def test_port_bound_nvgre(self):
+        self.agent._nvgre_enabled = True
+        self._check_port_bound_net_type(network_type=constants.TYPE_NVGRE)
+
         self.agent._nvgre_ops.bind_nvgre_port.assert_called_once_with(
             mock.sentinel.segmentation_id, mock.sentinel.vswitch_name,
             mock.sentinel.port_id)
@@ -478,19 +481,23 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
 
     @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
                        '_treat_vif_port')
-    def test_process_added_port(self, mock_treat_vif_port):
+    @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
+                       '_update_port_status_cache')
+    def test_process_added_port(self, mock_update_port_cache,
+                                mock_treat_vif_port):
         self.agent._added_ports = set()
         details = self._get_fake_port_details()
-
+        CONF.AGENT.enable_qos_extension = True
         self.agent._process_added_port(details)
 
+        self.agent._qos_ext.handle_port.assert_called_once_with(
+            self.agent.context, details)
         mock_treat_vif_port.assert_called_once_with(
             mock.sentinel.port_id, mock.sentinel.network_id,
             mock.sentinel.network_type, mock.sentinel.physical_network,
             mock.sentinel.segmentation_id, mock.sentinel.admin_state_up)
-        self.agent.plugin_rpc.update_device_up.assert_called_once_with(
-            self.agent.context, mock.sentinel.device,
-            self.agent.agent_id, self.agent._host)
+        mock_update_port_cache.assert_called_once_with(mock.sentinel.device,
+                                                       device_bound=True)
         self.assertNotIn(mock.sentinel.device, self.agent._added_ports)
 
     @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
@@ -520,7 +527,8 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
 
         self.assertIn(mock.sentinel.port_id, self.agent._added_ports)
 
-    def test_treat_devices_added_updates_known_port(self):
+    @mock.patch.object(hyperv_neutron_agent.eventlet, 'spawn_n')
+    def test_treat_devices_added_updates_known_port(self, mock_spawn):
         self.agent._added_ports = set([mock.sentinel.device])
         details = self._get_fake_port_details()
         attrs = {'get_devices_details_list.return_value': [details]}
@@ -528,7 +536,7 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
 
         self.agent._treat_devices_added()
 
-        self.agent._workers.submit.assert_called_once_with(
+        mock_spawn.assert_called_once_with(
             self.agent._process_added_port, details)
         self.assertNotIn(mock.sentinel.device, self.agent._added_ports)
 
@@ -540,39 +548,57 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
 
         self.agent._treat_devices_added()
 
-        self.assertFalse(self.agent._workers.submit.called)
         self.assertNotIn(mock.sentinel.port_id, self.agent._added_ports)
 
-    def test_treat_devices_removed_exception(self):
-        self.agent._removed_ports = set([mock.sentinel.port_id])
-        attrs = {'update_device_down.side_effect': Exception()}
-        self.agent.plugin_rpc.configure_mock(**attrs)
+    @mock.patch('eventlet.spawn_n')
+    def test_treat_devices_removed(self, mock_spawn):
+        mock_removed_ports = [mock.sentinel.port0, mock.sentinel.port1]
+        self.agent._removed_ports = set(mock_removed_ports)
+
         self.agent._treat_devices_removed()
 
-        self.agent.plugin_rpc.update_device_down.assert_called_once_with(
-            self.agent.context, mock.sentinel.port_id,
-            self.agent.agent_id, self.agent._host)
+        mock_spawn.assert_has_calls(
+            [mock.call(self.agent._process_removed_port, port)
+             for port in mock_removed_ports],
+            any_order=True)
+
+    @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
+                       '_port_unbound')
+    @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
+                       '_update_port_status_cache')
+    def test_process_removed_port_exception(self, mock_update_port_cache,
+                                            mock_port_unbound):
+        self.agent._removed_ports = set([mock.sentinel.port_id])
+
+        raised_exc = exception.NetworkingHyperVException
+        mock_port_unbound.side_effect = raised_exc
+
+        self.assertRaises(raised_exc,
+                          self.agent._process_removed_port,
+                          mock.sentinel.port_id)
+
+        mock_update_port_cache.assert_called_once_with(
+            mock.sentinel.port_id, device_bound=False)
         self.assertIn(mock.sentinel.port_id, self.agent._removed_ports)
 
-    def mock_treat_devices_removed(self, port_exists):
+    @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
+                       '_port_unbound')
+    @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
+                       '_update_port_status_cache')
+    def test_process_removed_port(self, mock_update_port_cache,
+                                  mock_port_unbound):
         self.agent._removed_ports = set([mock.sentinel.port_id])
-        details = dict(exists=port_exists)
-        attrs = {'update_device_down.return_value': details}
-        self.agent.plugin_rpc.configure_mock(**attrs)
-        with mock.patch.object(self.agent, '_port_unbound') as func:
-            self.agent._treat_devices_removed()
-        self.assertEqual(func.called, not port_exists)
-        self.assertEqual(
-            self.agent.sec_groups_agent.remove_devices_filter.called,
-            not port_exists)
+
+        self.agent._process_removed_port(mock.sentinel.port_id)
+
+        mock_update_port_cache.assert_called_once_with(
+            mock.sentinel.port_id, device_bound=False)
+        mock_port_unbound.assert_called_once_with(mock.sentinel.port_id,
+                                                  vnic_deleted=True)
+        self.agent.sec_groups_agent.remove_devices_filter(
+            [mock.sentinel.port_id])
         self.assertNotIn(mock.sentinel.port_id, self.agent._removed_ports)
 
-    def test_treat_devices_removed_unbinds_port(self):
-        self.mock_treat_devices_removed(False)
-
-    def test_treat_devices_removed_ignores_missing_port(self):
-        self.mock_treat_devices_removed(False)
-
     def test_process_added_port_event(self):
         self.agent._added_ports = set()
         self.agent._process_added_port_event(mock.sentinel.port_id)
@@ -583,32 +609,77 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
         self.agent._process_removed_port_event(mock.sentinel.port_id)
         self.assertIn(mock.sentinel.port_id, self.agent._removed_ports)
 
-    @mock.patch.object(hyperv_neutron_agent.threading, 'Thread')
-    def test_create_event_listeners(self, mock_Thread):
+    @mock.patch.object(hyperv_neutron_agent.eventlet, 'spawn_n')
+    def test_create_event_listeners(self, mock_spawn):
         self.agent._create_event_listeners()
 
         self.agent._utils.get_vnic_event_listener.assert_has_calls([
             mock.call(self.agent._utils.EVENT_TYPE_CREATE),
             mock.call(self.agent._utils.EVENT_TYPE_DELETE)])
         target = self.agent._utils.get_vnic_event_listener.return_value
-        calls = [mock.call(target=target,
-                           args=(self.agent._process_added_port_event, )),
-                 mock.call(target=target,
-                           args=(self.agent._process_removed_port_event, ))]
-        mock_Thread.assert_has_calls(calls, any_order=True)
-        self.assertEqual(2, mock_Thread.return_value.start.call_count)
+        calls = [mock.call(target,
+                           self.agent._process_added_port_event),
+                 mock.call(target,
+                           self.agent._process_removed_port_event)]
+        mock_spawn.assert_has_calls(calls, any_order=True)
 
-    def test_thread_pool_execution(self):
-        pool = futures.ThreadPoolExecutor(max_workers=3)
-        mock_fn = mock.MagicMock()
+    def test_update_port_status_cache_device_bound(self):
+        self.agent._bound_ports = set()
+        self.agent._unbound_ports = set([mock.sentinel.device])
 
-        for i in range(8):
-            pool.submit(mock_fn, mock.sentinel.parameter)
+        self.agent._update_port_status_cache(mock.sentinel.device,
+                                             device_bound=True)
 
-        # allow the threads to finish. one second is enough for a noop call.
-        time.sleep(1)
-        mock_fn.assert_has_calls([mock.call(mock.sentinel.parameter)] * 8)
+        self.assertIn(mock.sentinel.device, self.agent._bound_ports)
+        self.assertNotIn(mock.sentinel.device, self.agent._unbound_ports)
 
+    def test_update_port_status_cache_device_unbound(self):
+        self.agent._bound_ports = set([mock.sentinel.device])
+        self.agent._unbound_ports = set()
+
+        self.agent._update_port_status_cache(mock.sentinel.device,
+                                             device_bound=False)
+
+        self.assertIn(mock.sentinel.device, self.agent._unbound_ports)
+        self.assertNotIn(mock.sentinel.device, self.agent._bound_ports)
+
+    def test_notify_plugin_no_updates(self):
+        self.agent._bound_ports = set()
+        self.agent._unbound_ports = set()
+
+        self.agent._notify_plugin_on_port_updates()
+
+        self.assertFalse(self.agent.plugin_rpc.update_device_list.called)
+
+    def test_notify_plugin_on_port_updates(self):
+        bound_ports = [mock.sentinel.bound_port]
+        unbound_ports = [mock.sentinel.unbound_port]
+
+        new_bound_ports = set([mock.sentinel.new_bound_port])
+        new_unbound_ports = set([mock.sentinel.new_unbound_port])
+
+        self.agent._bound_ports = set(bound_ports)
+        self.agent._unbound_ports = set(unbound_ports)
+
+        def plugin_rpc_side_effect(*args, **kwargs):
+            # We test the scenario in which we're getting some new port
+            # updates during the plugin rpc call.
+            self.agent._bound_ports.update(new_bound_ports)
+            self.agent._unbound_ports.update(new_unbound_ports)
+
+        mock_update_device_list = self.agent.plugin_rpc.update_device_list
+        mock_update_device_list.side_effect = plugin_rpc_side_effect
+
+        self.agent._notify_plugin_on_port_updates()
+
+        mock_update_device_list.assert_called_once_with(
+            self.agent.context, bound_ports, unbound_ports,
+            self.agent.agent_id, self.agent._host)
+
+        self.assertEqual(new_bound_ports, self.agent._bound_ports)
+        self.assertEqual(new_unbound_ports, self.agent._unbound_ports)
+
+    @mock.patch('eventlet.spawn_n')
     @mock.patch('time.sleep')
     @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
                        '_port_enable_control_metrics')
@@ -617,7 +688,7 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
     @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
                        '_create_event_listeners')
     def test_daemon_loop(self, mock_create_listeners, mock_treat_dev_added,
-                         mock_port_enable_metrics, mock_sleep):
+                         mock_port_enable_metrics, mock_sleep, mock_spawn):
         self.agent._nvgre_enabled = True
         mock_port_enable_metrics.side_effect = KeyError
         mock_sleep.side_effect = KeyboardInterrupt
@@ -627,7 +698,12 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
         self.assertEqual(self.agent._utils.get_vnic_ids.return_value,
                          self.agent._added_ports)
         self.assertEqual(set(), self.agent._removed_ports)
+        self.assertEqual(set(), self.agent._bound_ports)
+        self.assertEqual(set(), self.agent._unbound_ports)
+
         mock_create_listeners.assert_called_once_with()
+        mock_spawn.assert_called_once_with(
+            self.agent._notify_plugin_on_port_updates)
         mock_treat_dev_added.assert_called_once_with()
         self.agent._nvgre_ops.refresh_nvgre_records.assert_called_once_with()
         mock_port_enable_metrics.assert_called_with()
