@@ -19,13 +19,12 @@ Unit tests for the Hyper-V Security Groups Driver.
 
 import mock
 from os_win import exceptions
-from os_win import utilsfactory
 
 from hyperv.neutron import security_groups_driver as sg_driver
 from hyperv.tests import base
 
 
-class SecurityGroupRuleTestHelper(base.BaseTestCase):
+class SecurityGroupRuleTestHelper(base.HyperVBaseTestCase):
     _FAKE_DIRECTION = 'egress'
     _FAKE_ETHERTYPE = 'IPv4'
     _FAKE_ETHERTYPE_IPV6 = 'IPv6'
@@ -68,9 +67,6 @@ class TestHyperVSecurityGroupsDriver(SecurityGroupRuleTestHelper):
 
     def setUp(self):
         super(TestHyperVSecurityGroupsDriver, self).setUp()
-        utilsfactory_patcher = mock.patch.object(utilsfactory, '_get_class')
-        utilsfactory_patcher.start()
-        self.addCleanup(utilsfactory_patcher.stop)
 
         self._driver = sg_driver.HyperVSecurityGroupsDriver()
         self._driver._utils = mock.MagicMock()
@@ -176,6 +172,15 @@ class TestHyperVSecurityGroupsDriver(SecurityGroupRuleTestHelper):
         self._driver._create_port_rules.assert_called_with(
             self._FAKE_ID, [fake_rule])
 
+    def test_prepare_port_filter_security_disabled(self):
+        mock_port = self._get_port()
+        mock_port.pop('port_security_enabled')
+
+        self._driver.prepare_port_filter(mock_port)
+
+        self.assertNotIn(mock_port['device'], self._driver._security_ports)
+        self.assertNotIn(mock_port['id'], self._driver._sec_group_rules)
+
     @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
                        '_generate_rules')
     def test_update_port_filter(self, mock_gen_rules):
@@ -220,6 +225,71 @@ class TestHyperVSecurityGroupsDriver(SecurityGroupRuleTestHelper):
         self.assertNotIn(new_mock_port['device'], self._driver._security_ports)
         mock_method.assert_called_once_with(new_mock_port)
 
+    def test_update_port_filter_security_disabled(self):
+        mock_port = self._get_port()
+        mock_port['port_security_enabled'] = False
+
+        self._driver.update_port_filter(mock_port)
+
+        self.assertFalse(self._driver._utils.remove_all_security_rules.called)
+        self.assertNotIn(mock_port['device'], self._driver._security_ports)
+        self.assertNotIn(mock_port['id'], self._driver._sec_group_rules)
+
+    def test_update_port_filter_security_disabled_existing_rules(self):
+        mock_port = self._get_port()
+        mock_port.pop('port_security_enabled')
+        self._driver._sec_group_rules[mock_port['id']] = mock.ANY
+
+        self._driver.update_port_filter(mock_port)
+
+        self._driver._utils.remove_all_security_rules.assert_called_once_with(
+            mock_port['id'])
+        self.assertNotIn(mock_port['device'], self._driver._security_ports)
+        self.assertNotIn(mock_port['id'], self._driver._sec_group_rules)
+
+    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
+                       '_generate_rules')
+    def test_update_port_filter_existing_wildcard_rules(self, mock_gen_rules):
+        mock_port = self._get_port()
+        new_mock_port = self._get_port()
+        new_mock_port['id'] += '2'
+        new_mock_port['security_group_rules'][0]['ethertype'] += "2"
+
+        fake_rule_new = self._create_security_rule(direction='egress',
+                                                   protocol='udp')
+        fake_wildcard_rule_new = self._create_security_rule(protocol='ANY',
+                                                            direction='egress')
+        fake_expanded_rules = [
+            self._create_security_rule(direction='egress', protocol='tcp'),
+            self._create_security_rule(direction='egress', protocol='udp'),
+            self._create_security_rule(direction='egress', protocol='icmp')]
+        self._driver._sg_gen.expand_wildcard_rules.return_value = (
+            fake_expanded_rules)
+
+        mock_gen_rules.return_value = {
+            new_mock_port['id']: [fake_rule_new, fake_wildcard_rule_new]}
+
+        self._driver._security_ports[mock_port['device']] = mock_port
+        self._driver._sec_group_rules[new_mock_port['id']] = [
+            self._create_security_rule(direction='egress', protocol='icmp')]
+
+        filtered_new_rules = [new_mock_port['security_group_rules'][0],
+                              fake_wildcard_rule_new]
+        filtered_remove_rules = mock_port['security_group_rules']
+
+        self._driver._create_port_rules = mock.MagicMock()
+        self._driver._remove_port_rules = mock.MagicMock()
+        self._driver.update_port_filter(new_mock_port)
+
+        self._driver._sg_gen.expand_wildcard_rules.assert_called_once_with(
+            [fake_rule_new, fake_wildcard_rule_new])
+        self._driver._remove_port_rules.assert_called_once_with(
+            mock_port['id'], filtered_remove_rules)
+        self._driver._create_port_rules.assert_called_once_with(
+            new_mock_port['id'], filtered_new_rules)
+        self.assertEqual(new_mock_port,
+                         self._driver._security_ports[new_mock_port['device']])
+
     def test_remove_port_filter(self):
         mock_port = self._get_port()
         mock_rule = mock.MagicMock()
@@ -340,7 +410,8 @@ class TestHyperVSecurityGroupsDriver(SecurityGroupRuleTestHelper):
         return {
             'device': self._FAKE_DEVICE,
             'id': self._FAKE_ID,
-            'security_group_rules': [mock.MagicMock()]
+            'security_group_rules': [mock.MagicMock()],
+            'port_security_enabled': True
         }
 
 
@@ -451,6 +522,38 @@ class SecurityGroupRuleGeneratorR2TestCase(SecurityGroupRuleR2BaseTestCase):
 
         self.assertEqual([new_rule], add_rules)
 
+    def test_expand_wildcard_rules(self):
+        egress_wildcard_rule = self._create_security_rule(
+            protocol='ANY',
+            direction='egress')
+        ingress_wildcard_rule = self._create_security_rule(
+            protocol='ANY',
+            direction='ingress')
+        normal_rule = self._create_security_rule()
+        rules = [egress_wildcard_rule, ingress_wildcard_rule, normal_rule]
+
+        actual_expanded_rules = self.sg_gen.expand_wildcard_rules(rules)
+
+        expanded_rules = []
+        for proto in sg_driver.ACL_PROP_MAP['protocol'].keys():
+            expanded_rules.extend(
+                [self._create_security_rule(protocol=proto,
+                                            direction='egress'),
+                 self._create_security_rule(protocol=proto,
+                                            direction='ingress')])
+        diff_expanded_rules = [r for r in expanded_rules
+                               if r not in actual_expanded_rules]
+        self.assertEqual([], diff_expanded_rules)
+
+    def test_expand_no_wildcard_rules(self):
+        normal_rule = self._create_security_rule(direction='egress')
+        another_normal_rule = self._create_security_rule(direction='ingress')
+
+        actual_expanded_rules = self.sg_gen.expand_wildcard_rules(
+            [normal_rule, another_normal_rule])
+
+        self.assertEqual([], actual_expanded_rules)
+
     def test_get_rule_port_range(self):
         rule = self._create_security_rule()
         expected = '%s-%s' % (self._FAKE_PORT_MIN, self._FAKE_PORT_MAX)
